@file:Suppress("DEPRECATION")

package com.mediamain.sdk_flutter.util

import android.Manifest
import android.annotation.SuppressLint
import android.content.Context
import android.content.pm.PackageManager
import android.net.ConnectivityManager
import android.net.NetworkInfo
import android.os.Build
import android.telephony.ServiceState
import android.telephony.SubscriptionManager
import android.telephony.TelephonyManager
import android.text.TextUtils
import androidx.core.app.ActivityCompat

object NetworkTypeUtils {
    fun getNetworkType(context: Context): NetworkType {
        if (isEthernet(context)) {
            return NetworkType.NETWORK_ETHERNET
        }
        if (ActivityCompat.checkSelfPermission(context, Manifest.permission.ACCESS_NETWORK_STATE) == PackageManager.PERMISSION_GRANTED) {
            val info: NetworkInfo? = getActiveNetworkInfo(context)
            if (info != null && info.isAvailable) {
                if (info.type == ConnectivityManager.TYPE_WIFI) {
                    return NetworkType.NETWORK_WIFI
                } else if (info.type == ConnectivityManager.TYPE_MOBILE) {
                    when (info.subtype) {
                        TelephonyManager.NETWORK_TYPE_GSM, TelephonyManager.NETWORK_TYPE_GPRS,
                        TelephonyManager.NETWORK_TYPE_CDMA,TelephonyManager.NETWORK_TYPE_EDGE,
                        TelephonyManager.NETWORK_TYPE_1xRTT, TelephonyManager.NETWORK_TYPE_IDEN -> return NetworkType.NETWORK_2G
                        TelephonyManager.NETWORK_TYPE_TD_SCDMA, TelephonyManager.NETWORK_TYPE_EVDO_A,
                        TelephonyManager.NETWORK_TYPE_UMTS, TelephonyManager.NETWORK_TYPE_EVDO_0,
                        TelephonyManager.NETWORK_TYPE_HSDPA, TelephonyManager.NETWORK_TYPE_HSUPA,
                        TelephonyManager.NETWORK_TYPE_HSPA, TelephonyManager.NETWORK_TYPE_EVDO_B,
                        TelephonyManager.NETWORK_TYPE_EHRPD, TelephonyManager.NETWORK_TYPE_HSPAP -> return NetworkType.NETWORK_3G
                        TelephonyManager.NETWORK_TYPE_IWLAN, TelephonyManager.NETWORK_TYPE_LTE -> return NetworkType.NETWORK_4G
                        TelephonyManager.NETWORK_TYPE_NR -> return NetworkType.NETWORK_5G
                        else -> {
                            val subtypeName = info.subtypeName
                            if (subtypeName.equals("TD-SCDMA", ignoreCase = true)
                                    || subtypeName.equals("WCDMA", ignoreCase = true)
                                    || subtypeName.equals("CDMA2000", ignoreCase = true)) {
                                return NetworkType.NETWORK_3G
                            }
                        }
                    }
                }
            }
        }
        try {
            if (NetworkUtil.getNetWorkType(context) == TelephonyManager.NETWORK_TYPE_NR) {
                return NetworkType.NETWORK_5G
            }
        } catch (ignore: Exception) {
        }
        return NetworkType.NETWORK_UNKNOWN
    }

    private fun isEthernet(context: Context): Boolean {
        if (ActivityCompat.checkSelfPermission(context, Manifest.permission.ACCESS_NETWORK_STATE) == PackageManager.PERMISSION_GRANTED) {
            val cm = context.getSystemService(Context.CONNECTIVITY_SERVICE) as ConnectivityManager
                    ?: return false
            val info = cm.getNetworkInfo(ConnectivityManager.TYPE_ETHERNET) ?: return false
            val state = info.state ?: return false
            return state == NetworkInfo.State.CONNECTED || state == NetworkInfo.State.CONNECTING
        }
        return false
    }

    enum class NetworkType {
        NETWORK_ETHERNET, NETWORK_WIFI, NETWORK_5G, NETWORK_4G, NETWORK_3G, NETWORK_2G, NETWORK_UNKNOWN, NETWORK_NO
    }

    object NetworkUtil {
        /**
         * get the network type
         *
         * @param ctx Context
         * @return networktype
         */
        fun getNetWorkType(ctx: Context): Int {
            var networkType = 0
            val tm = ctx.getSystemService(Context.TELEPHONY_SERVICE) as TelephonyManager
            try {
                val defaultDataSubId = subId
                if (defaultDataSubId == -1) {
                    networkType = tm.networkType
                } else {
                    try {
                        val dataNetworkType = TelephonyManager::class.java
                                .getDeclaredMethod("getDataNetworkType", *arrayOf<Class<*>?>(Int::class.javaPrimitiveType))
                        dataNetworkType.isAccessible = true
                        networkType = dataNetworkType.invoke(tm, defaultDataSubId) as Int
                    } catch (t: Throwable) {
                        networkType = tm.networkType
                    }
                }
            } catch (t: Throwable) {
                // do nothing
            }
            if (networkType == TelephonyManager.NETWORK_TYPE_LTE) {
                networkType = adjustNetworkType(ctx, networkType)
            }
            return networkType
        }

        /**
         * get the 5G network type
         *
         * @param ctx Context
         * @param networkTypeFromSys this method can be call only when networkTypeFromSys = 13(LET)
         * @return correct network type
         */
        @SuppressLint("PrivateApi")
        private fun adjustNetworkType(ctx: Context, networkTypeFromSys: Int): Int {
            var networkType = 0
            if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.O
                    && ctx.checkSelfPermission(Manifest.permission.READ_PHONE_STATE)
                    == PackageManager.PERMISSION_GRANTED) {
                try {
                    val tm = ctx
                            .getSystemService(Context.TELEPHONY_SERVICE) as TelephonyManager
                    var ss: ServiceState?
                    val defaultDataSubId = subId
                    if (defaultDataSubId == -1) {
                        ss = tm.serviceState
                    } else {
                        try {
                            val infTm = TelephonyManager::class.java
                            val method = infTm
                                    .getDeclaredMethod("getServiceStateForSubscriber",
                                            *arrayOf<Class<*>?>(Int::class.javaPrimitiveType))
                            method.isAccessible = true
                            ss = method.invoke(tm, defaultDataSubId) as ServiceState
                        } catch (t: Throwable) {
                            ss = tm.serviceState
                        }
                    }
                    if (ss != null && isServiceStateFiveGAvailable(ss.toString())) {
                        networkType = TelephonyManager.NETWORK_TYPE_NR
                    }
                } catch (e: Exception) {
                    // do nothing
                }
            }
            return networkType
        }

        /**
         * get data sub id
         *
         * @return subId
         */
        private val subId: Int
            private get() {
                var defaultDataSubId = -1
                if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.N) {
                    defaultDataSubId = SubscriptionManager.getDefaultDataSubscriptionId()
                }
                return defaultDataSubId
            }

        /**
         * check the service state str is 5G
         *
         * @param ss services state str
         * @return true if is 5G
         */
        private fun isServiceStateFiveGAvailable(ss: String): Boolean {
            var available = false
            if (!TextUtils.isEmpty(ss)
                    && (ss.contains("nrState=NOT_RESTRICTED")
                            || ss.contains("nrState=CONNECTED"))) {
                available = true
            }
            return available
        }
    }

    private fun getActiveNetworkInfo(ctx: Context): NetworkInfo? {
        if (ActivityCompat.checkSelfPermission(ctx, Manifest.permission.ACCESS_NETWORK_STATE) == PackageManager.PERMISSION_GRANTED) {
            val cm = ctx.getSystemService(Context.CONNECTIVITY_SERVICE) as ConnectivityManager
            return cm.activeNetworkInfo
        }
        return null
    }
}