//
//  TATClient.m
//  TATMediaSDK
//
//  Created by wuleslie on 2019/12/4.
//  Copyright © 2019 wuleslie. All rights reserved.
//
#import "TATClient.h"
#import <UIKit/UIKit.h>
#import <AdSupport/AdSupport.h>
#import "TATCocoaSecurity.h"
#import <sys/utsname.h>
#import <sys/sysctl.h>
#import <ifaddrs.h>
#import <arpa/inet.h>
#import <SystemConfiguration/CaptiveNetwork.h>
#import <CommonCrypto/CommonCrypto.h>
#import <CoreTelephony/CTTelephonyNetworkInfo.h>
#import <CoreTelephony/CTCarrier.h>
#import "TATKeychainHelper.h"
#import "TATLocationManager.h"
#import "HAIJIANPPNetworkReachabilityManager.h"
#import <mach/mach.h>
#import "TATStorageHelper.h"
#import "NSString+TATBase.h"
#import "NSDate+TATFormatter.h"

static TATClient *current_client = nil;

// version code要对应哦
static NSString *const kSDKVersion = @"3.1.1.0";
static NSInteger const kSDKVersionCode = 3110;

@interface TATClient ()

@property(nonatomic, readwrite, copy) NSString *appKey;
@property(nonatomic, readwrite, copy) NSString *appSecret;
@property(nonatomic, readwrite, assign) BOOL isInitialized;
@property(nonatomic, readwrite, copy) NSString *deviceIP;
@property(nonatomic, strong) NSLock *lock;

@end

@implementation TATClient

+ (TATClient *)current {
    static dispatch_once_t onceToken;
    dispatch_once(&onceToken, ^{
        if (!current_client) {
            current_client = [[TATClient alloc] init];
        }
    });
    return current_client;
}

+ (instancetype)registerWithSecret:(NSString *)secret appKey:(NSString *)appKey {
    //NSParameterAssert(secret);
    //NSParameterAssert(appKey);
    [TATClient current].appKey = appKey;
    [TATClient current].appSecret = secret;
    [TATClient current].isInitialized = YES;
    return current_client;
}

- (instancetype)init {
    self = [super init];
    if (self) {
        [[HAIJIANPPNetworkReachabilityManager sharedManager] startMonitoring];
    }
    return self;
}

- (void)getCurrentLocation {
    [TATLocationManager getLocationWithSuccess:^(CLLocationDegrees latitude, CLLocationDegrees longitude) {
        current_client.longitude = [@(longitude) stringValue];
        current_client.latitude = [@(latitude) stringValue];
        [TATLocationManager stop];
    } failure:^(NSError * _Nullable error) {
        [TATLocationManager stop];
    }];
}

+ (NSString *)uuidString {
    CFUUIDRef uuid_ref = CFUUIDCreate(NULL);
    CFStringRef uuid_string_ref = CFUUIDCreateString(NULL, uuid_ref);
    NSString *uuid = [NSString stringWithString:(__bridge NSString *)uuid_string_ref];
    CFRelease(uuid_ref);
    CFRelease(uuid_string_ref);
    return [uuid lowercaseString];
}

#pragma mark - getters

- (NSString *)systemName {
    return @"iOS";
}

- (NSString *)systemVersion {
    return [UIDevice currentDevice].systemVersion;
}

- (NSString *)idfa {
    if ([self.blockList containsObject:NSStringFromSelector(_cmd)]) {
        return @"";
    }
    return [[TATCocoaSecurity md5:[self idfa_original]] hexLower];
}

- (NSString *)idfa_original {
    if ([self.blockList containsObject:NSStringFromSelector(_cmd)]) {
        return @"";
    }
    return [TATStorageHelper idfa_original];
}

- (NSString *)bundleIdentifier {
    if ([self.blockList containsObject:NSStringFromSelector(_cmd)]) {
        return @"";
    }
    return [[NSBundle mainBundle] infoDictionary][@"CFBundleIdentifier"];
}

- (NSString *)appBuild {
    if ([self.blockList containsObject:NSStringFromSelector(_cmd)]) {
        return @"";
    }
    return [[NSBundle mainBundle] infoDictionary][@"CFBundleVersion"];
}

- (NSString *)appVersion {
    if ([self.blockList containsObject:NSStringFromSelector(_cmd)]) {
        return @"";
    }
    return [[NSBundle mainBundle] infoDictionary][@"CFBundleShortVersionString"];
}

- (NSString *)screenSize {
    if ([self.blockList containsObject:NSStringFromSelector(_cmd)]) {
        return @"";
    }
    return [NSString stringWithFormat:@"%.f X %.f", [UIScreen mainScreen].bounds.size.width, [UIScreen mainScreen].bounds.size.height];
}

- (NSString *)screenResolution {
    if ([self.blockList containsObject:NSStringFromSelector(_cmd)]) {
        return @"";
    }
    CGFloat scale = [UIScreen mainScreen].scale;
    return [NSString stringWithFormat:@"%.f X %.f", [UIScreen mainScreen].bounds.size.width * scale, [UIScreen mainScreen].bounds.size.height * scale];
}

- (NSString *)scale {
    if ([self.blockList containsObject:NSStringFromSelector(_cmd)]) {
        return @"";
    }
    return [NSString stringWithFormat:@"%f", [UIScreen mainScreen].scale];
}

- (NSString *)appPath {
    if ([self.blockList containsObject:NSStringFromSelector(_cmd)]) {
        return @"";
    }
    return [NSBundle mainBundle].bundlePath;
}

- (NSString *)appName {
    if ([self.blockList containsObject:NSStringFromSelector(_cmd)]) {
        return @"";
    }
    return [[NSBundle mainBundle] infoDictionary][@"CFBundleName"];
}

- (NSString *)deviceId {
    if ([self.blockList containsObject:NSStringFromSelector(_cmd)]) {
        return @"";
    }
    return [TATStorageHelper idfa_original]; // 要求统一用idfa
}

- (NSString *)tcid {
    NSString *tcid = [TATStorageHelper tcid];
    if (![tcid tat_isNotEmpty]) {
        NSString *uuid = [TATClient uuidString];
        NSString *mixedId = [NSString stringWithFormat:@"%lld%@", [NSDate tat_currentTimestamp], uuid];
        NSString *md5Id = [[TATCocoaSecurity md5:mixedId] hexLower];
        if (md5Id.length < 10) { // exception
            return @"";
        }
        NSString *prefixCode = @"1020";
        tcid = [md5Id stringByReplacingCharactersInRange:NSMakeRange(0, 4) withString:prefixCode];
        [TATStorageHelper cacheTcid:tcid];
    }
    return tcid;
}

- (NSString *)authStatus {
    if ([[self idfa_original] tat_isNotEmpty]) {
        return @"1";
    }
    return @"0";
}

- (NSInteger)sdkVersionCode {
    return kSDKVersionCode;
}

- (NSString *)networkType {
    if ([self.blockList containsObject:NSStringFromSelector(_cmd)]) {
        return @"";
    }
    HAIJIANPPNetworkReachabilityStatus status = [HAIJIANPPNetworkReachabilityManager sharedManager].networkReachabilityStatus;
    switch (status) {
        case HAIJIANPPNetworkReachabilityStatusNotReachable:
            return @"Not Reachable";
        case HAIJIANPPNetworkReachabilityStatusReachableViaWiFi:
            return @"wifi";
        case HAIJIANPPNetworkReachabilityStatusReachableViaWWAN:
            return [self WWANType];
        case HAIJIANPPNetworkReachabilityStatusUnknown:
            return @"unknown";
    }
}

- (NSString *)WWANType {
    if ([self.blockList containsObject:NSStringFromSelector(_cmd)]) {
        return @"";
    }
    CTTelephonyNetworkInfo *netinfo = [[CTTelephonyNetworkInfo alloc] init];
    NSString *radioTech = netinfo.currentRadioAccessTechnology;
    NSArray *type2GArray = @[CTRadioAccessTechnologyEdge,
                             CTRadioAccessTechnologyGPRS,
                             CTRadioAccessTechnologyCDMA1x];
    NSArray *type3GArray = @[CTRadioAccessTechnologyHSDPA,
                             CTRadioAccessTechnologyWCDMA,
                             CTRadioAccessTechnologyHSUPA,
                             CTRadioAccessTechnologyCDMAEVDORev0,
                             CTRadioAccessTechnologyCDMAEVDORevA,
                             CTRadioAccessTechnologyCDMAEVDORevB,
                             CTRadioAccessTechnologyeHRPD];
    if ([type2GArray containsObject:radioTech]) {
        return @"2G";
    } else if ([type3GArray containsObject:radioTech]) {
        return @"3G";
    } else if ([radioTech isEqualToString:CTRadioAccessTechnologyLTE]) {
        return @"4G";
    } else if (@available(iOS 14.1, *)) {
        if ([radioTech isEqualToString:CTRadioAccessTechnologyNRNSA]) {
            return @"5G NSA";
        } else if ([radioTech isEqualToString:CTRadioAccessTechnologyNR]) {
            return @"5G";
        }
    }
    return @"unknown";
}

- (NSString *)dataFromType {
    return @"0"; // 目前iOS没有原生插屏，均为SDK
}

@end
