//
//  TATStorageHelper.m
//  TATMediaSDK
//
//  Created by wuleslie on 2020/6/3.
//  Copyright © 2020 wuleslie. All rights reserved.
//

#import "TATStorageHelper.h"
#import "NSString+TATBase.h"
#import "TATKeychainHelper.h"
#import <AdSupport/AdSupport.h>
#import <AdSupport/ASIdentifierManager.h>

#define TAT_EMPTY_IDFA @"00000000-0000-0000-0000-000000000000"

static NSString *const kTATTcidKey = @"tat_tcid_key";
static NSString *const kTATIDFAOriginalKey = @"tat_idfa_original_key";
static NSString *const kTATSimulateIDFAKey = @"tat_simulate_idfa_key";
static NSString *const kTATSystemBootTimeKey = @"tat_system_boot_time_key";


@implementation TATStorageHelper

+ (id)valueForKey:(NSString *)key {
    if (![key tat_isNotEmpty]) {
        return nil;
    }
    NSString *prefixKey = [TATStorageHelper keyWithPrefix:key];
    return [[NSUserDefaults standardUserDefaults] objectForKey:prefixKey];
}

+ (void)setObject:(id)object forKey:(NSString *)key {
    if (![key tat_isNotEmpty]) {
        return;
    }
    NSString *prefixKey = [TATStorageHelper keyWithPrefix:key];
    [[NSUserDefaults standardUserDefaults] setObject:object forKey:prefixKey];
    [[NSUserDefaults standardUserDefaults] synchronize];
}

+ (NSInteger)exposureTimesForSlotId:(NSString *)slotId {
    if (![slotId tat_isNotEmpty]) {
        return 0;
    }
    NSString *prefixKey = [TATStorageHelper keyWithPrefix:slotId];
    return [[NSUserDefaults standardUserDefaults] integerForKey:prefixKey];
}

+ (void)exposeSlotId:(NSString *)slotId {
    if (![slotId tat_isNotEmpty]) {
        return;
    }
    NSInteger times = [TATStorageHelper exposureTimesForSlotId: slotId];
    times++;
    NSString *prefixKey = [TATStorageHelper keyWithPrefix:slotId];
    [[NSUserDefaults standardUserDefaults] setInteger:times forKey:prefixKey];
    [[NSUserDefaults standardUserDefaults] synchronize];
}

+ (NSString *)tcid {
    NSString *tcid = [TATKeychainHelper dataForService:kTATTcidKey];
    if ([tcid tat_isNotEmpty]) {
        return tcid;
    }
    tcid = [[NSUserDefaults standardUserDefaults] stringForKey:kTATTcidKey];
    if ([tcid tat_isNotEmpty]) {
        //  从keychain中未取出tcid，但从UserDefaults中取到了UserDefaults，此种情况下需要将tcid存储到keychain
        [self cacheTcid:tcid];
    }
    return tcid;
}

+ (void)cacheTcid:(NSString *)tcid {
    if (![tcid tat_isNotEmpty]) {
        return;
    }
    [TATKeychainHelper save:kTATTcidKey data:tcid];
}

+ (NSString *)idfa_original {
    NSString *idfa_original = [TATKeychainHelper dataForService:kTATIDFAOriginalKey];
    NSString *emptyIDFA = TAT_EMPTY_IDFA;
    BOOL isEmptyIDFA;
    NSString *currentIdfa_original;
    //
    currentIdfa_original = [[[ASIdentifierManager sharedManager] advertisingIdentifier] UUIDString];
    // 用户限制了广告追踪
    isEmptyIDFA = ![ASIdentifierManager.sharedManager isAdvertisingTrackingEnabled];
            
    if (!isEmptyIDFA) {
        // 获取到的IDFA不为空
        if (![idfa_original tat_isNotEmpty] || ![currentIdfa_original isEqualToString:idfa_original]) {
            
            // 存keychain中储的IDFA为空，或者当前获取到的IDFA不等于之前keychain中存储的IDFA
            idfa_original = currentIdfa_original;
            [self cacheIdfa_original:idfa_original];
        }
    } else if (![idfa_original tat_isNotEmpty]) {
        // 获取到的IDFA为空, 且存储在keychain的IDFA也为空
        idfa_original = emptyIDFA;
    }
    // 默认都没有进入所有ifelse语句的话说明当前获取IDFA为空且存储的IDFA不为空
    return idfa_original;
}

+ (void)cacheIdfa_original:(NSString *)idfa_original {
    if (![idfa_original tat_isNotEmpty]) {
        return;
    }
    [TATKeychainHelper save:kTATIDFAOriginalKey data:idfa_original];
}

#pragma mark - Private

+ (NSString *)keyWithPrefix:(NSString *)originalKey {
    return [NSString stringWithFormat:@"tat_%@", originalKey];
}

@end
